@@ -1,4 +1,4 @@
-# coding: utf-8
+# codingr utf-8
 # Copyright 2009 Alexandre Fiori
 # https://github.com/fiorix/txredisapi
 #
@@ -36,12 +36,16 @@ import hashlib
 from twisted.internet import defer
 from twisted.internet import protocol
 from twisted.internet import reactor
-from twisted.internet import task
 from twisted.protocols import basic
 from twisted.protocols import policies
 from twisted.python import log
 from twisted.python.failure import Failure
 
+try:
+    import hiredis
+except ImportError:
+    hiredis = None
+
 
 class RedisError(Exception):
     pass
@@ -80,7 +84,10 @@ def list_or_args(command, keys, args):
     try:
         iter(keys)
         if isinstance(keys, (str, unicode)):
-            raise TypeError
+            keys = [keys]
+            if not oldapi:
+                return keys
+            oldapi = True
     except TypeError:
         oldapi = True
         keys = [keys]
@@ -189,7 +196,7 @@ class LineReceiver(protocol.Protocol, basic._PauseableMixin):
         return self.transport.loseConnection()
 
 
-class RedisProtocol(LineReceiver, policies.TimeoutMixin):
+class BaseRedisProtocol(LineReceiver, policies.TimeoutMixin):
     """
     Redis client protocol.
     """
@@ -355,25 +362,32 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
         """
         el = None
         if data is not None:
-            if data and data[0] in _NUM_FIRST_CHARS:  # Most likely a number
-                try:
-                    el = int(data) if data.find('.') == -1 else float(data)
-                except ValueError:
-                    pass
-
-            if el is None:
-                el = data
-                if self.charset is not None:
-                    try:
-                        el = data.decode(self.charset)
-                    except UnicodeDecodeError:
-                        pass
+            el = self.tryConvertData(data)
 
         if self.multi_bulk.pending or self.multi_bulk.items:
             self.handleMultiBulkElement(el)
         else:
             self.replyReceived(el)
 
+    def tryConvertData(self, data):
+        if not isinstance(data, str):
+            return data
+        el = None
+        if data and data[0] in _NUM_FIRST_CHARS:  # Most likely a number
+            try:
+                el = int(data) if data.find('.') == -1 else float(data)
+            except ValueError:
+                pass
+
+        if el is None:
+            el = data
+            if self.charset is not None:
+                try:
+                    el = data.decode(self.charset)
+                except UnicodeDecodeError:
+                    pass
+        return el
+
     def handleMultiBulkElement(self, element):
         self.multi_bulk.append(element)
 
@@ -393,27 +407,32 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
             reply = self.multi_bulk.items
             self.multi_bulk = MultiBulkStorage()
 
-            if self.inTransaction and reply is not None: # watch or multi has been called
-                if self.transactions > 0:
-                    self.transactions -= len(reply)      # multi: this must be an exec [commit] reply
-                if self.transactions == 0:
-                    self.commit_cc()
-                if self.inTransaction:                   # watch but no multi: process the reply as usual
-                    f = self.post_proc[1:]
-                    if len(f) == 1 and callable(f[0]):
-                        reply = f[0](reply)
-                else:                                    # multi: this must be an exec reply
-                    tmp = []
-                    for f, v in zip(self.post_proc[1:], reply):
-                        if callable(f):
-                            tmp.append(f(v))
-                        else:
-                            tmp.append(v)
-                        reply = tmp
-                self.post_proc = []
+            reply = self.handleTransactionData(reply)
 
             self.replyReceived(reply)
 
+    def handleTransactionData(self, reply):
+        if self.inTransaction and isinstance(reply, list):  # watch or multi has been called
+            if self.transactions > 0:
+                # multi: this must be an exec [commit] reply
+                self.transactions -= len(reply)
+            if self.transactions == 0:
+                self.commit_cc()
+            if self.inTransaction:  # watch but no multi: process the reply as usual
+                f = self.post_proc[1:]
+                if len(f) == 1 and callable(f[0]):
+                    reply = f[0](reply)
+            else:  # multi: this must be an exec reply
+                tmp = []
+                for f, v in zip(self.post_proc[1:], reply):
+                    if callable(f):
+                        tmp.append(f(v))
+                    else:
+                        tmp.append(v)
+                    reply = tmp
+            self.post_proc = []
+        return reply
+
     def replyReceived(self, reply):
         """
         Complete reply received and ready to be pushed to the requesting
@@ -610,23 +629,24 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
         return self.execute_command("FLUSHALL")
 
     # Commands operating on string values
-    def set(self, key, value, preserve=False, getset=False):
+    def set(self, key, value, expire=None, pexpire=None,
+            only_if_not_exists=False, only_if_exists=False):
         """
         Set a key to a string value
         """
-        if preserve:
-            warnings.warn(DeprecationWarning(
-                "preserve option to 'set' is deprecated, "
-                "use redis.setnx() instead"))
-            return self.setnx(key, value)
-
-        if getset:
-            warnings.warn(DeprecationWarning(
-                "getset option to 'set' is deprecated, "
-                "use redis.getset() instead"))
-            return self.getset(key, value)
-
-        return self.execute_command("SET", key, value)
+        args = []
+        if expire is not None:
+            args.extend(("EX", expire))
+        if pexpire is not None:
+            args.extend(("PX", pexpire))
+        if only_if_not_exists and only_if_exists:
+            raise RedisError("only_if_not_exists and only_if_exists "
+                             "cannot be true simultaneously")
+        if only_if_not_exists:
+            args.append("NX")
+        if only_if_exists:
+            args.append("XX")
+        return self.execute_command("SET", key, value, *args)
 
     def get(self, key):
         """
@@ -871,7 +891,7 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
         keys.append(timeout)
         return self.execute_command("BRPOP", *keys)
 
-    def brpoplpush(self, source, destination, timeout = 0):
+    def brpoplpush(self, source, destination, timeout=0):
         """
         Pop a value from a list, push it to another list and return
         it; or block until one is available.
@@ -1394,7 +1414,8 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
     @defer.inlineCallbacks
     def execute_pipeline(self):
         if not self.pipelining:
-            raise RedisError("Not currently pipelining commands, please use pipeline() first")
+            err = "Not currently pipelining commands, please use pipeline() first"
+            raise RedisError(err)
 
         # Flush all the commands at once to redis. Wait for all replies
         # to come back using a deferred list.
@@ -1514,8 +1535,8 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
         keys_and_args = tuple(keys) + tuple(args)
         r = self.execute_command("EVALSHA",
                                  sha1_hash, n,
-                                *keys_and_args).addErrback(self._evalsha_errback,
-                                                   sha1_hash)
+                                 *keys_and_args).addErrback(self._evalsha_errback,
+                                                            sha1_hash)
         if sha1_hash not in self.script_hashes:
             r.addCallback(self._eval_success, sha1_hash)
         return r
@@ -1556,6 +1577,49 @@ class RedisProtocol(LineReceiver, policies.TimeoutMixin):
     def script_load(self, script):
         return self.execute_command("SCRIPT",  "LOAD", script)
 
+    # Redis 2.8.9 HyperLogLog commands
+    def pfadd(self, key, elements, *args):
+        elements = list_or_args("pfadd", elements, args)
+        return self.execute_command("PFADD", key, *elements)
+
+    def pfcount(self, keys, *args):
+        keys = list_or_args("pfcount", keys, args)
+        return self.execute_command("PFCOUNT", *keys)
+
+    def pfmerge(self, destKey, sourceKeys, *args):
+        sourceKeys = list_or_args("pfmerge", sourceKeys, args)
+        return self.execute_command("PFMERGE", destKey, *sourceKeys)
+
+
+class HiredisProtocol(BaseRedisProtocol):
+    def __init__(self, *args, **kwargs):
+        BaseRedisProtocol.__init__(self, *args, **kwargs)
+        self._reader = hiredis.Reader(protocolError=InvalidData,
+                                      replyError=ResponseError)
+
+    def dataReceived(self, data, unpause=False):
+        self.resetTimeout()
+        if data:
+            self._reader.feed(data)
+        res = self._reader.gets()
+        while res is not False:
+            if isinstance(res, basestring):
+                res = self.tryConvertData(res)
+            elif isinstance(res, list):
+                res = map(self.tryConvertData, res)
+            if res == "QUEUED":
+                self.transactions += 1
+            else:
+                res = self.handleTransactionData(res)
+
+            self.replyReceived(res)
+            res = self._reader.gets()
+
+if hiredis is not None:
+    RedisProtocol = HiredisProtocol
+else:
+    RedisProtocol = BaseRedisProtocol
+
 
 class MonitorProtocol(RedisProtocol):
     """
@@ -1619,10 +1683,6 @@ class ConnectionHandler(object):
         self._factory = factory
         self._connected = factory.deferred
 
-    def _wait_pool_cleanup(self, deferred):
-        if self._factory.size == 0:
-            deferred.callback(True)
-
     def disconnect(self):
         self._factory.continueTrying = 0
         for conn in self._factory.pool:
@@ -1631,15 +1691,12 @@ class ConnectionHandler(object):
             except:
                 pass
 
-        d = defer.Deferred()
-        t = task.LoopingCall(self._wait_pool_cleanup, d)
-        d.addCallback(lambda ign: t.stop())
-        t.start(.5)
-        return d
+        return self._factory.waitForEmptyPool()
 
     def __getattr__(self, method):
         def wrapper(*args, **kwargs):
             d = self._factory.getConnection()
+
             def callback(connection):
                 protocol_method = getattr(connection, method)
                 try:
@@ -1775,7 +1832,7 @@ class HashRing(object):
     def get_node(self, key):
         n, i = self.get_node_pos(key)
         return n
-    #self.get_node_pos(key)[0]
+    # self.get_node_pos(key)[0]
 
     def get_node_pos(self, key):
         if len(self.ring) == 0:
@@ -1923,6 +1980,7 @@ class RedisFactory(protocol.ReconnectingClientFactory):
         self.deferred = defer.Deferred()
         self.handler = handler(self)
         self.connectionQueue = defer.DeferredQueue()
+        self._waitingForEmptyPool = set()
 
     def buildProtocol(self, addr):
         if hasattr(self, 'charset'):
@@ -1948,6 +2006,25 @@ class RedisFactory(protocol.ReconnectingClientFactory):
             log.msg("Could not remove connection from pool: %s" % str(e))
 
         self.size = len(self.pool)
+        if not self.size and self._waitingForEmptyPool:
+            deferreds = self._waitingForEmptyPool
+            self._waitingForEmptyPool = set()
+            for d in deferreds:
+                d.callback(None)
+
+    def _cancelWaitForEmptyPool(self, deferred):
+        self._waitingForEmptyPool.discard(deferred)
+        deferred.errback(defer.CancelledError())
+
+    def waitForEmptyPool(self):
+        """
+        Returns a Deferred which fires when the pool size has reached 0.
+        """
+        if not self.size:
+            return defer.succeed(None)
+        d = defer.Deferred(self._cancelWaitForEmptyPool)
+        self._waitingForEmptyPool.add(d)
+        return d
 
     def connectionError(self, why):
         if self.deferred:
@@ -2048,7 +2125,8 @@ def ShardedConnection(hosts, dbid=None, reconnect=True, charset="utf-8", passwor
     return makeShardedConnection(hosts, dbid, 1, reconnect, False, charset, password)
 
 
-def lazyShardedConnection(hosts, dbid=None, reconnect=True, charset="utf-8", password=None):
+def lazyShardedConnection(hosts, dbid=None, reconnect=True, charset="utf-8",
+                          password=None):
     return makeShardedConnection(hosts, dbid, 1, reconnect, True, charset, password)
 
 
@@ -2116,7 +2194,8 @@ def lazyUnixConnectionPool(path="/tmp/redis.sock", dbid=None, poolsize=10,
     return makeUnixConnection(path, dbid, poolsize, reconnect, True, charset, password)
 
 
-def ShardedUnixConnection(paths, dbid=None, reconnect=True, charset="utf-8", password=None):
+def ShardedUnixConnection(paths, dbid=None, reconnect=True, charset="utf-8",
+                          password=None):
     return makeShardedUnixConnection(paths, dbid, 1, reconnect, False, charset, password)
 
 
